/* $Workfile: T8pSs7.c $       $Revision: 8 $      $Modtime: 9/30/04 3:45p $  */
/*******************************************************************************

  Odin TeleSystems Inc.

  OTX Telecom Adapter Driver

  @doc PublicDemo

  @module

    T8PSS7.C

    A simple example console application demonstrating the use of OTX Driver.

    The applications works with the following board combinations:

      Thor-8-PCI-Plus + Vidar-55x4-ASM,

    The sample application provides a simple console interface which utilizes
    simple keystrokes as commands.

    It configures 4 E1T1 devices on the board. The E1T1 spans should be
    looped with an external loop cable (e.g. Li0<->Li1, Li2<->Li3, etc).
    
    It then configures 31 HDLC channels (senders and receivers) per DSP with 
    FISU/LSSU filtering enabled. 

    There is a sender and a receiver thread per HDLC channel. The sender thread 
    sends random MSU frames at a certain load. The receiver thread receives the 
    MSU and verifies that it contains the correct payload (i.e. what the sender
    thread sent).

    Press 'h' or '?' for help. 'q' to quit.

*******************************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <assert.h>
#include <time.h>

#include "OsaStd.h"
#include "OsaCon.h"
#include "OsaThrd.h"
//#include "OtxGeneric.h"
#include "OtxErr.h"
#include "OtxBrd.h"
#include "OtxDef.h"
#include "OtxType.h"
#include "OtxDrv.h"
#include "OtxTone.h"
#include "OtxDsp.h"
#include "OtxHdlc.h"
#include "OtxT1E1.h"
#include "OtxData.h"

#define DEMO_TASK_REF 1L               // Task Reference for a non-blocking call
#define DEMO_HDLC_RX_BUFFER_SIZE 2048
#define DEMO_MAX_FRAME_LEN 57
#define DEMO_NO_OF_DEVS 2
#define DEMO_DEFAULT_LOG_FILE "T8pSs7.log"
#define DEMO_DEFAULT_BOARD_NO 0

typedef struct DemoEnvT {
  char                  *pszInputStr;
  OTX_BOOL                bExitOnError;
  char                  *logFileName;
  OtxPhysicalDeviceTypeE nBoardType;
  char                   cwd[OTX_MAX_PATH];
  int                    coDsp;
  int                    coLi;
  int                    coHdlcPerLi;
  OtxT1E1LiModeE         eT1E1Mode;    // T1 or E1 mode
  OTX_BOOL               bSs7FisuFiltering;
  OTX_BOOL               bHdlcBitInversion;
  int                    boardNo;
  OTX_BOOL               bUseDspDebugger;
  OTX_BOOL               bUseSenders;
  OTX_BOOL               bUseReceivers;
  OTX_BOOL               b56kbpsChannel;
  int                    nTimeSlots;
  OtxT1LineCodeE         eT1LineCode;
  OtxT1FrameFormatE      eT1FrameFormat;
  OtxE1LineCodeE         eE1LineCode;
  OtxE1FrameFormatE      eE1FrameFormat;
  OtxDspSpmE             eProgramModule;      // Standard Program Module to be loaded in the DSP (from the Lib)
  OtxBrdClockSourceE     eClkSource;
} DemoEnvT;

void *DemoMsuTransmitFunction(
  void *lpParameter   // thread data
);

/*******************************************************************************
 Function Prototypes
*******************************************************************************/
static void DemoCreateEventQueue(void);
static void DemoOpenPhysicalDevices(void);
static void DemoConfigureE1T1Devices(void);
static void DemoEnablePhysicalDevices(void);
static void DemoCreateLogicalDevices(void);
static void DemoConnectDevices(void);
static void DemoInitializeLogicalDevices(void);
static void DemoEnableLogicalDevices(void);
static void DemoInitT1E1ConfigData(OUT OtxE1ConfigOptionsS *pE1CfgData, OUT OtxT1ConfigOptionsS *pT1CfgData);
static void DemoPrintEvent(IN OtxEventDataS *pEventData);

static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP);
static void DemoHelpPanel(void);
static void DemoLogf(char *formatStr, ...);
static void DemoLogf(char *formatStr, ...);
static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT otxRc);
static void DemoExitQuestion(void);

static OTX_INT32 DemoProcessUserCommand(char commandChar);

/*******************************************************************************
 Global Data
*******************************************************************************/
OTX_EVENT   g_hNotificationEvent = OTX_INVALID_EVENT_VALUE;
OTX_HANDLE  g_hEventQueue  = 0;
FILE       *g_pDemoLogFile = NULL;
DemoEnvT    g_demoEnv;

int g_counter=0;

OsaMutexS g_demoLogfCritSec;

struct {
  OTX_HANDLE hThorController0;
  OTX_HANDLE hQuadE1T1[4]; 
  OTX_HANDLE hVidarAsmController0; 
  OTX_HANDLE hLi[16];
  OTX_HANDLE hDsp[4];
} g_hPhysDevs;

struct {
  OTX_INT32 nDev;

  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;
  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;

  OTX_BOOL bActive;
  OTX_BOOL bVerbose;

  OTX_INT32  nLi;
  OTX_INT32  nDsp;
  OTX_INT32  nTs;

  OtxHdlcSS7FisuS tFisu;
  OtxHdlcSS7ShortLssuS tLssu;

  OTX_INT32 coMsusToSend;
  OTX_INT32 coMsusSent;
  OTX_INT32 nMinSleepTimeBetweenHdlcFrames;
  OTX_INT32 nLoadFactor;

  OTX_INT32 nByteCount;
  OTX_INT32 nBusyCount;
  clock_t startTime, stopTime;

} g_sqDspHdlcSender[DEMO_NO_OF_DEVS];

struct {
  OTX_INT32 nDev;

  OTX_HANDLE hPipe;
  OTX_HANDLE hHandle;
  OTX_HANDLE hEventQueue;
  OTX_EVENT  hNotificationEvent;

  OTX_BOOL bActive;
  OTX_BOOL bVerbose;

  OTX_INT32  nLi;
  OTX_INT32  nDsp;
  OTX_INT32  nTs;

  OtxHdlcSS7FisuS tFisu;
  OtxHdlcSS7ShortLssuS tLssu;
  OTX_UINT8 nPrevBsn;

  OTX_INT32 coMsusReceived;
  OTX_INT32 coMsusToReceive;

} g_sqDspHdlcReceiver[DEMO_NO_OF_DEVS];


struct {
  OtxE1ConfigOptionsS e1CfgData;
  OtxT1ConfigOptionsS t1CfgData;
} g_T1E1CfgData;


/*******************************************************************************
*******************************************************************************/


/*******************************************************************************

  @func
    Initalize the global data for the Hdlc Sender and Receiver devices

*******************************************************************************/
void DemoInitializeGlobalData(
  void
)
{
  OTX_INT32 i;

  for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
    g_sqDspHdlcSender[i].nDev = g_sqDspHdlcReceiver[i].nDev = i;

    g_sqDspHdlcSender[i].nTs  = g_sqDspHdlcReceiver[i].nTs  = (i%32);
    g_sqDspHdlcSender[i].nLi  = g_sqDspHdlcReceiver[i].nLi  = (i/32);
    g_sqDspHdlcSender[i].nDsp = g_sqDspHdlcReceiver[i].nDsp = (i/32);

    g_sqDspHdlcSender[i].bActive = g_sqDspHdlcReceiver[i].bActive = OTX_FALSE;
    g_sqDspHdlcSender[i].bVerbose = g_sqDspHdlcReceiver[i].bVerbose = OTX_FALSE;

    g_sqDspHdlcSender[i].coMsusToSend = 200000;
    g_sqDspHdlcSender[i].coMsusSent = 0;
    g_sqDspHdlcSender[i].nMinSleepTimeBetweenHdlcFrames = 0; // in milliseconds
    g_sqDspHdlcSender[i].nLoadFactor = 1; // Increasing this value this increase the load. Lowest value is 1.
    g_sqDspHdlcSender[i].tFisu.bib = 1;
    g_sqDspHdlcSender[i].tFisu.fib = 1;
    g_sqDspHdlcSender[i].tFisu.fsn = 0x7F;
    g_sqDspHdlcSender[i].tFisu.bsn = 0x7F;


    g_sqDspHdlcReceiver[i].coMsusToReceive = 200000;
    g_sqDspHdlcReceiver[i].coMsusReceived = 0;
    g_sqDspHdlcReceiver[i].tFisu.bib = 1;
    g_sqDspHdlcReceiver[i].tFisu.fib = 1;
    g_sqDspHdlcReceiver[i].tFisu.fsn = 0x7F;
    g_sqDspHdlcReceiver[i].tFisu.bsn = 0x7F;

    g_sqDspHdlcReceiver[i].nPrevBsn   = 0x7F;


  }
}




/*******************************************************************************

  @func

  Create the event queue to be used by the driver to inform the application
  about events, such as line status changes, or detected DMTF digits.

*******************************************************************************/
void DemoCreateEventQueue(void)
{
  OTX_RESULT result;

  // Create a notification Event
  result = OtxDrvCreateEvent(&g_hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);

  // Create Event Queue for driver notifications
  result = OtxDrvCreateEventQueue(128, g_hNotificationEvent, &g_hEventQueue); DemoCheckRc(g_hEventQueue, result);

  // Enable Event Queue
  result = OtxDrvEnable(g_hEventQueue, OTX_TASK_SYNC,0); DemoCheckRc(g_hEventQueue, result);
}



/*******************************************************************************

  @func
    Open the Physical Devices on the Thor board

  @comm
    Calls function <f OtxDrvCreatePhysicalDevice> returning a handle to a
    Device Object representing a physical device, which can be the
    entire board or an integrated circuit (chip) on the board. The number of
    physical devices is set per board and cannot change dynamically.

*******************************************************************************/
void DemoOpenPhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Connect to the Physical Board Controller
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_PCI:
    DemoLogf("Opening OTX_DEVICE_THOR_PCI board\n");
    break;
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    DemoLogf("Opening Thor-8-PCI-Plus board\n");
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    DemoLogf("Opening Gimle-16-PCI-Plus board\n");
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }
  result = OtxDrvOpenPhysicalDevice(0, g_demoEnv.nBoardType, g_demoEnv.boardNo, g_demoEnv.boardNo, g_hEventQueue, &g_hPhysDevs.hThorController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    {
      OTX_INT32 nQuadE1T1; // Only used for Thor-8-PCI and Thor-8-PCI-Plus

      // The E1T1 devices are hosted by quad E1T1 devices. Open those two devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < 8; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    {
      OTX_INT32 nQuadE1T1; // Only used for Thor-8-PCI and Thor-8-PCI-Plus

      // The E1T1 devices are hosted by quad E1T1 devices. Open those four devices.
      for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
        DemoLogf("Open OTX_DEVICE_QUAD_T1E1 #%d\n", nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, OTX_DEVICE_QUAD_T1E1, nQuadE1T1, nQuadE1T1, g_hEventQueue, &g_hPhysDevs.hQuadE1T1[nQuadE1T1]); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      }
      
      // Get handles to T1/E1 line interface devices 
      for (i = 0; i < 16; i++) {
        nQuadE1T1 = i/4;
        DemoLogf("Open OTX_DEVICE_LI_T1E1 #%d (hosted by OTX_DEVICE_QUAD_T1E1 #%d)\n", i, nQuadE1T1);
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_DEVICE_LI_T1E1, i%4, i, g_hEventQueue, &g_hPhysDevs.hLi[i]); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  // Open physical DSP devices
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_8_PCI_PLUS: 
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    {
      OTX_UINT32   eDevType;
      OTX_ITERATOR itrChildTypes;
      OtxDrvIteratorBegin(&itrChildTypes);
      while (OTX_SUCCESS(OtxDrvEnumDeviceType(g_hPhysDevs.hThorController0, &itrChildTypes, (OTX_INT32 *)&eDevType))) {  
        OTX_BOOL bOpenAsm = OTX_FALSE;
        switch (eDevType) {
        case OTX_DEVICE_VIDAR_5x_ASM:
        case OTX_DEVICE_VIDAR_5X4_ASM_CST:
        case OTX_DEVICE_VIDAR_5X4_ASM_V90:
        case OTX_DEVICE_VIDAR_5X4_ASM_EX:
        case OTX_DEVICE_VIDAR_5X4_ASM_PRO:
          DemoLogf("Asm Type (%d) not supported by this demo program. Only OTX_DEVICE_VIDAR_55x4_ASM is supported.\n", eDevType);
          exit(1);
          break;
        case OTX_DEVICE_VIDAR_55x4_ASM:
          DemoLogf("Opening Vidar-55x4-ASM board\n");
          bOpenAsm = OTX_TRUE;
          break;
        }
        if (bOpenAsm) {
          // Get a handle to to the ASM Daughter Board
          result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hThorController0, (OtxPhysicalDeviceTypeE) eDevType, 0, 0, g_hEventQueue, &g_hPhysDevs.hVidarAsmController0); DemoCheckRc(g_hPhysDevs.hThorController0, result);
        }
      }
      
      for (i = 0; i < g_demoEnv.coDsp; i++) {
        result = OtxDrvOpenPhysicalDevice(g_hPhysDevs.hVidarAsmController0, OTX_DEVICE_DSP, i, i, g_hEventQueue, &g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
      }
    }
    break;
  default:
    DemoLogf("Board Type (%d) not supported by this demo program\n", g_demoEnv.nBoardType);
    exit(1);
    break;
  }

  // Get DSP Highway Speed
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    OTX_UINT32  nDspHighwaySpeed;  // DSP Highway speed in kpbs
    OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[i], OTX_ATTR_DSP_HIGHWAY_SPEED, &nDspHighwaySpeed); 
    DemoLogf("DSP#%d highway speed is %d kbps\n", i, nDspHighwaySpeed);
  }

	// Load DSP Program
  for (i = 0; i < g_demoEnv.coDsp; i++) {
    switch (g_demoEnv.eProgramModule) {
    case OTX_DSP_SPM_OTXHDLC:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXHDLC\n", i);
      break;
    case OTX_DSP_SPM_OTXHDLCEX:
      DemoLogf("DSP#%d: Loading DSP program OTX_DSP_SPM_OTXHDLCEX\n", i);
      break;
    default:
      DemoLogf("Unsupported DSP program module (%d). Exiting.\n", g_demoEnv.eProgramModule);
      OsaConsoleExit();
      OtxDrvReleaseLib();
      exit(1);
    }
#if 1
    result = OtxDspRunSpm(g_hPhysDevs.hDsp[i], g_demoEnv.eProgramModule); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
#else
    DemoLogf("Loading from file\n");
    //result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], "OtxSpm2.out", "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    //result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], "OtxSpm2c5410a.out", "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    //result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], "OtxHdlc5410a.out", "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    result = OtxDspRunProgram(g_hPhysDevs.hDsp[i], "OtxHdlcEx5416.out", "_c_int00"); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
#endif
  }

  if (g_demoEnv.bUseDspDebugger) {
    DemoLogf("Please load and run DSP program %s to DSP#%d-%d in your DSP Debugger\n", "OtxHdlc.out", 0, g_demoEnv.coDsp-1);
    DemoLogf("Press any key when all the DSP programs are running...\n");
    OsaConsoleGetCh();
    
    // At this point we need to tell the host to restart the programs 
    // (to issue the initial IO commands to the DSP again)
    for (i = 0; i < g_demoEnv.coDsp; i++) {
      DemoLogf("Restarting program on DSP#%d\n", i);
      result = OtxDspRestartProgram(g_hPhysDevs.hDsp[i]); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
    }
    // At this point the DSP should be in the same state as they were right after OtxDspRunProgram() was called
  }


}



/*******************************************************************************

  @func
    Configure E1/T1 Devices

  @comm The clock source of the board(s) must set by a call 
  to <f OtxBrdSetClocks> prior to calling this function.

*******************************************************************************/
void DemoConfigureE1T1Devices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  // Initialize the T1/E1 Line Interface Devices

  // Set default configuration data
  DemoInitT1E1ConfigData( &(g_T1E1CfgData.e1CfgData), &(g_T1E1CfgData.t1CfgData));

  // Configure Transceivers
  for (i = 0; i < g_demoEnv.coLi; i++) {

    // Set termination for the T1/E1 Front End
    result = OtxT1E1SetLineTermination(g_hPhysDevs.hLi[i], OTX_T1E1_TERM_100_OHM); DemoCheckRc(g_hPhysDevs.hLi[i], result);

    if (g_demoEnv.eT1E1Mode == OTX_T1E1_MODE_T1) {
      // T1 mode
      result = OtxT1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.t1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    } else {
      // E1 mode
      result = OtxE1SetConfiguration(g_hPhysDevs.hLi[i], &(g_T1E1CfgData.e1CfgData));  DemoCheckRc(g_hPhysDevs.hLi[i], result);
    }
    result = OtxT1E1SetLiMode(g_hPhysDevs.hLi[i], g_demoEnv.eT1E1Mode); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

}

/*******************************************************************************

  @func
    Enable Configured Physical Devices

*******************************************************************************/
void DemoEnablePhysicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  result = OtxDrvEnable(g_hPhysDevs.hThorController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  if (g_hPhysDevs.hVidarAsmController0) {
    result = OtxDrvEnable(g_hPhysDevs.hVidarAsmController0,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hVidarAsmController0, result);
  }

  // If board type is Thor-8-PCI and Thor-8-PCI-Plus then enable Quad E1T1 device
  switch (g_demoEnv.nBoardType) {
  case OTX_DEVICE_THOR_8_PCI_PLUS:
    {
      OTX_INT32 nQuadE1T1; 
      
      // Enable Quad E1T1 devices
      for (nQuadE1T1 = 0; nQuadE1T1 < 2; nQuadE1T1++) {
        DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
        result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  case OTX_DEVICE_GIMLE_16_PCI_PLUS:
    {
      OTX_INT32 nQuadE1T1; 
      
      // Enable Quad E1T1 devices
      for (nQuadE1T1 = 0; nQuadE1T1 < 4; nQuadE1T1++) {
        DemoLogf("Enable OTX_DEVICE_QUAD_T1E1 #%d device\n", nQuadE1T1);
        result = OtxDrvEnable(g_hPhysDevs.hQuadE1T1[nQuadE1T1], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hQuadE1T1[nQuadE1T1], result);
      }
    }
    break;
  default:
    // Nothing needed
    break;
  }

  for (i = 0; i < g_demoEnv.coLi; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hLi[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hLi[i], result);
  }

  for (i = 0; i < g_demoEnv.coDsp; i++) {
    result = OtxDrvEnable(g_hPhysDevs.hDsp[i], OTX_TASK_SYNC, 5000L); DemoCheckRc(g_hPhysDevs.hDsp[i], result);
  }
}



/*******************************************************************************

  @func
    Create the needed Logical Devices

  @comm
    The logical devices are differenciated from the phyiscal devices by the fact
    the number of logical devices is not set. More of them can be requested
    by the user and allocated by the system if needed.

*******************************************************************************/
void DemoCreateLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;


  for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
    if (g_demoEnv.bUseSenders) {
      // Create a notification Event and Event Queue for the Sender Device
      result = OtxDrvCreateEvent(&g_sqDspHdlcSender[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
      result = OtxDrvCreateEventQueue(128, g_sqDspHdlcSender[i].hNotificationEvent, &g_sqDspHdlcSender[i].hEventQueue); DemoCheckRc(g_sqDspHdlcSender[i].hEventQueue, result);
      result = OtxDrvEnable(g_sqDspHdlcSender[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcSender[i].hEventQueue, result);
      
      // Creat the sender device
      result = OtxDrvCreateLogicalDevice(
        g_hPhysDevs.hDsp[g_sqDspHdlcSender[i].nDsp], 
        OTX_LDEVICE_HDLC_SENDER,   
        g_sqDspHdlcSender[i].hEventQueue, 
        i, 
        &g_sqDspHdlcSender[i].hHandle); 
      DemoCheckRc(g_hPhysDevs.hDsp[g_sqDspHdlcSender[i].nDsp], result);
    }

    if (g_demoEnv.bUseReceivers) {
      // Create a notification Event and Event Queue for the Receiver Device
      result = OtxDrvCreateEvent(&g_sqDspHdlcReceiver[i].hNotificationEvent); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
      result = OtxDrvCreateEventQueue(128, g_sqDspHdlcReceiver[i].hNotificationEvent, &g_sqDspHdlcReceiver[i].hEventQueue); DemoCheckRc(g_sqDspHdlcReceiver[i].hEventQueue, result);
      result = OtxDrvEnable(g_sqDspHdlcReceiver[i].hEventQueue, OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcReceiver[i].hEventQueue, result);
      
      result = OtxDrvCreateLogicalDevice(
        g_hPhysDevs.hDsp[g_sqDspHdlcReceiver[i].nDsp], 
        OTX_LDEVICE_HDLC_RECEIVER, 
        g_sqDspHdlcReceiver[i].hEventQueue, 
        i, 
        &g_sqDspHdlcReceiver[i].hHandle); 
      DemoCheckRc(g_hPhysDevs.hDsp[g_sqDspHdlcReceiver[i].nDsp], result);
    }
  }
  
}


/*******************************************************************************

  @func
    Create the needed Pipe Devices and connect the physical and logical devices
    to pipes.

  @comm
    A Pipe Device is a logical entity representing a serial bit-stream (a subset of a
    Highway) with a data rate between 8 kbit/s and 2.048Mbit/s. A pipe is a characterized
    bu its capacity (in 8 kbit/s increments). The user can create a pipe with
    the driver function <f OtxDrvCreatePipe>. A pipe is connected to physical or
    logical devices and mapped to specific time-slots with the driver function
    <f OtxDrvConnectPipe>.

*******************************************************************************/
void DemoConnectDevices(
  void
)
{
  OTX_INT32 i;
  OTX_RESULT result;
  OTX_UINT8 sqTimeSlotMask[32];
  OTX_INT32 nLi;
  OTX_INT32 nDsp;
  OTX_INT32 nTs;
  OTX_INT32 nCapacity;

  // 64kbps or 56kbps HDLC capacity
  if (g_demoEnv.b56kbpsChannel) {
    DemoLogf("Configuring one 56kbps HDLC channel per Li\n");
    nCapacity = 56;
  } else {
    DemoLogf("Configuring one 64kbps HDLC channel per Li\n");
    nCapacity = 64;
  }

  if (g_demoEnv.bUseSenders) {
    for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
      
      nLi  = g_sqDspHdlcSender[i].nLi;
      nDsp = g_sqDspHdlcSender[i].nDsp;
      nTs  = g_sqDspHdlcSender[i].nTs;
      
      if (nTs == 0) continue; // Skip TS0 (sync)
      
      result = OtxDrvCreatePipe(nCapacity, &g_sqDspHdlcSender[i].hPipe); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
      
      // Connect pipes to the E1/T1 device
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      if (g_demoEnv.b56kbpsChannel) {
        sqTimeSlotMask[nTs] = 0xFE; // Timeslot on the T1/E1 span
      } else {
        sqTimeSlotMask[nTs] = 0xFF; // Timeslot on the T1/E1 span
      }
      
      result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_OUTPUT, g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
      result = OtxDrvConnectPipe(g_sqDspHdlcSender[i].hPipe,   OTX_PIPE_INPUT,  g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_sqDspHdlcSender[i].hPipe, result);
      
      // Make the following connections in transmit direction:
      // HDLC Sender   -----> Li
      result = OtxDrvConnectLogicalDevice(g_sqDspHdlcSender[i].hHandle, -1, g_sqDspHdlcSender[i].hPipe, -1); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
      
      DemoLogf("Hdlc Sender[%d]   connected (Dsp#%d, Li#%d) \n", i, g_sqDspHdlcSender[i].nDsp, g_sqDspHdlcSender[i].nLi);
    }
  }
    

  if (g_demoEnv.bUseReceivers) {
    for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
      
      nLi  = g_sqDspHdlcReceiver[i].nLi;
      nDsp = g_sqDspHdlcReceiver[i].nDsp;
      nTs  = g_sqDspHdlcReceiver[i].nTs;
      
      if (nTs == 0) continue; // Skip TS0 (sync)
      
      result = OtxDrvCreatePipe(nCapacity, &g_sqDspHdlcReceiver[i].hPipe); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
      
      // Connect pipes to the E1/T1 device
      memset(sqTimeSlotMask, 0, 32 * sizeof(OTX_UINT8));
      if (g_demoEnv.b56kbpsChannel) {
        sqTimeSlotMask[nTs] = 0xFE; // Timeslot on the T1/E1 span
      } else {
        sqTimeSlotMask[nTs] = 0xFF; // Timeslot on the T1/E1 span
      }
      
      result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_INPUT,  g_hPhysDevs.hLi[nLi],  -1, sqTimeSlotMask, 32); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
      result = OtxDrvConnectPipe(g_sqDspHdlcReceiver[i].hPipe, OTX_PIPE_OUTPUT, g_hPhysDevs.hDsp[nDsp], -1, NULL, 0); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
      
      // Create the following Connections in receive direction:
      //  Li  ------> HDLC Receiver
      result = OtxDrvConnectLogicalDevice(g_sqDspHdlcReceiver[i].hPipe, -1, g_sqDspHdlcReceiver[i].hHandle,       -1); DemoCheckRc(g_sqDspHdlcReceiver[i].hPipe, result);
      
      DemoLogf("Hdlc Receiver[%d] connected (Dsp#%d, Li#%d) \n", i, g_sqDspHdlcReceiver[i].nDsp, g_sqDspHdlcReceiver[i].nLi);
    }
  }
    

}


/*******************************************************************************

  @func
    Initialize the logical device. This function must be called after
    the devices are created and connected.

*******************************************************************************/
void DemoInitializeLogicalDevices(
  void
)
{
  OTX_RESULT result;
  OTX_INT32  i;

  for (i = 0; i < DEMO_NO_OF_DEVS; i++) {

    if (g_sqDspHdlcSender[i].nTs == 0) continue; // Skip TS0 (sync)


    if (g_demoEnv.bUseSenders) {
      
#if 1
      // Change the IRP queue size for the HDLC Sender 
      // (which is the number of HDLC frames that the driver will 
      // buffer before returning OTX_E_BUSY)
      // The default IRP queue size is 10 frames.
      result = OtxDrvSetAttributeValueUint32(
        g_sqDspHdlcSender[i].hHandle, 
        OTX_ATTR_DEVICE_IRP_QUEUE_LIMIT, 
        100); 
      DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
#endif
      
      
      // Initialize the Hdlc Sender 
      result = OtxHdlcSenderInit(
        g_sqDspHdlcSender[i].hHandle, // Handle to logical HDLC Filter Device
        OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
        1,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
        OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
        g_demoEnv.bHdlcBitInversion, // If OTX_TRUE then all bits are inverted when transmitted and before received
        OTX_TRUE,                  // If OTX_TRUE the HDLC device will append the CRC to the frame to be transmitted. If OTX_FALSE, the CRC is generated with software
        OTX_HDLC_TR_NA             // HDLC Transmission Rate for V.110 and X.30 modes
        ); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);          
    }
    
    if (g_demoEnv.bUseReceivers) {
      // Initialize the Hdlc Receiver 
      result = OtxHdlcReceiverInit(
        g_sqDspHdlcReceiver[i].hHandle, // Handle to logical HDLC Filter Device
        OTX_HDLC_FFT_FLAGS,        // Fill character(s) to be send between HDLC frames
        1,                         // Minimum No of interframe time characters between 2 HDLC frames (1=shared flags, 2=non-shared) Min value is 1. Max is 1024.
        OTX_HDLC_CRC16,            // Cyclic Redundancy Check (CRC) mode: 16-bit, 32-bit, or none.
        g_demoEnv.bHdlcBitInversion, // If OTX_TRUE then all bits are inverted when transmitted and before received
        OTX_TRUE,                  // If OTX_TRUE the HDLC device will remove the CRC from the received frame, if OTX_FALSE the CRC is passed on to the receiving software
        OTX_HDLC_TR_NA,            // HDLC Transmission Rate for V.110 and X.30 modes
        OTX_TRUE                  // If OTX_TRUE the HDLC device will generate an event for every received frame, if OTX_FALSE then the HDLC device will only generate an event when one or more frames have been received
        ); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);          
      
      if (g_demoEnv.bSs7FisuFiltering) {
        //if (i == 1) continue;
        result = OtxHdlcSS7SetFilter(g_sqDspHdlcReceiver[i].hHandle, OTX_HDLC_SS7_FILTER_FISU | OTX_HDLC_SS7_FILTER_SHORT_LSSU | OTX_HDLC_SS7_FILTER_LONG_LSSU); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);
      }
    }
    
  }
}



/*******************************************************************************

  @func
    Enable Configured Logical Devices

*******************************************************************************/
void DemoEnableLogicalDevices(void)
{
  OTX_RESULT result;
  OTX_INT32  i;

  if (g_demoEnv.bUseSenders) {
    for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
      if (g_sqDspHdlcSender[i].nTs == 0) continue; // Skip TS0 (sync)
      result = OtxDrvEnable(g_sqDspHdlcSender[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
    }
  }

  if (g_demoEnv.bUseReceivers) {
    for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
      if (g_sqDspHdlcReceiver[i].nTs == 0) continue; // Skip TS0 (sync)
      result = OtxDrvEnable(g_sqDspHdlcReceiver[i].hHandle,  OTX_TASK_SYNC, 5000L); DemoCheckRc(g_sqDspHdlcReceiver[i].hHandle, result);
    }
  }
  
}


/*******************************************************************************

  @func
    Initialize the T1/E1 Configuration Data Structures passed to the T1/E1
    Line Interface physical devices

*******************************************************************************/
void DemoInitT1E1ConfigData(
  OUT OtxE1ConfigOptionsS *pE1CfgData,
  OUT OtxT1ConfigOptionsS *pT1CfgData
)
{
  // Initialize E1 Configuration Data
  pE1CfgData->eTxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eRxLineCode                = g_demoEnv.eE1LineCode;
  pE1CfgData->eTxFrameFormat             = g_demoEnv.eE1FrameFormat;
  pE1CfgData->eRxFrameFormat             = g_demoEnv.eE1FrameFormat;
#if 1
  pE1CfgData->eAisDetection              = OTX_E1_AIS_ETS300233;
#else
  pE1CfgData->eAisDetection              = OTX_E1_AIS_G775;
#endif
  pE1CfgData->bfSiBits                   = 0x3;
  pE1CfgData->bfSaBits                   = 0x1F;
  pE1CfgData->bExtendedHDB3errorDetect   = OTX_FALSE;
  pE1CfgData->bAutoRegainMultiframe      = OTX_TRUE;
  pE1CfgData->bAutoResync                = OTX_TRUE;
  pE1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pE1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pE1CfgData->nLosSensitivity            = 128;
  pE1CfgData->nLosRecovery               = 128;
  pE1CfgData->nLineLength                = 10;
  pE1CfgData->bHighTxPower               = OTX_FALSE;
  pE1CfgData->bReceiveEqualizer          = OTX_FALSE;

  // Initialize T1 Configuration Data
  pT1CfgData->eSignalingMode             = OTX_T1_SM_CCS;
  pT1CfgData->eTxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eRxLineCode                = g_demoEnv.eT1LineCode;
  pT1CfgData->eFrameFormat               = g_demoEnv.eT1FrameFormat;
  pT1CfgData->bCRC6Enable                = OTX_TRUE;
  pT1CfgData->eTxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->eRxRemoteAlarmFormat       = OTX_T1_RA_YELLOW_A;
  pT1CfgData->bAutoResync                = OTX_TRUE;
  pT1CfgData->bAutoCrcBitInversion       = OTX_FALSE;
  pT1CfgData->eLfaSensitivity            = OTX_T1_LFA_2_OUT_OF_4;
  pT1CfgData->bAutoRemoteAlarm           = OTX_FALSE;
  pT1CfgData->nLosSensitivity            = 128;
  pT1CfgData->nLosRecovery               = 128;
  pT1CfgData->nLineLength                = 10;
  pT1CfgData->bHighTxPower               = OTX_FALSE;
  pT1CfgData->bReceiveEqualizer          = OTX_TRUE;
  pT1CfgData->bfClearChannels            = 0x000000;
}



/*******************************************************************************
  @func
    Prints a timestamp in a MM-DD hh:mm:ss:mmm format, where

    MM  = month
    DD  = day (of month)
    hh  = hour of day
    mm  = minutes of day
    ss  = seconds of day
    mmm = milliseconds

*******************************************************************************/
void DemoPrintTimeStamp(
  OTX_DATETIME nTimeStamp
)
{
  OTX_DATETIME nTimeStampMs;
  OTX_DATETIME nTimeStampSec;
  struct Osa_tm *ptCurrentDayAndTime;
  Osa_time_t nCurrentTimeSecs;
  
  // Subtract number of 100ns ticks between Jan 1, 1970 and Jan 1, 1600
  
  #if defined(__linux__)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 1164447360 * 100000000;  
  #elif defined(WIN32)
    nTimeStamp.nUint64 = nTimeStamp.nUint64 - 116444736000000000;  
  #endif

  
  // Convert to seconds since Jan 1, 1970 (so we can use this time with OsaStdLocalTime())
  nTimeStampSec.nUint64 = nTimeStamp.nUint64 / 10000000;
  
  // Calculate number of milliseconds (since number of seconds is not a very accurate way to report timestamp)
  nTimeStampMs.nUint64 = nTimeStamp.nUint64 / 10000 - (nTimeStampSec.nUint64 * 1000);
  
  nCurrentTimeSecs = nTimeStampSec.sUint32.nLow;

  ptCurrentDayAndTime = OsaStdLocalTime(&nCurrentTimeSecs);
  
  if (ptCurrentDayAndTime) {
    
    DemoLogf("%.2d-%.2d %.2d:%.2d:%.2d.%.3d ", 
      ptCurrentDayAndTime->tm_mon + 1,
      ptCurrentDayAndTime->tm_mday,
      ptCurrentDayAndTime->tm_hour,
      ptCurrentDayAndTime->tm_min, 
      ptCurrentDayAndTime->tm_sec,
      nTimeStampMs.sUint32.nLow
      );
  } else {
    //DemoLogf("Error: Invalid timestamp (before Jan 1, 1970)\n");
  }
  
}




/*******************************************************************************
  @func
    Idle Function that check for Events from the OTX Driver
*******************************************************************************/
void  *DemoIdleFunction(
  void *lpParameter   // thread data
)
{
  OtxEventDataS eventData;
  OTX_RESULT result;


  while (1) {
    OtxDrvWaitForSingleEvent(g_hNotificationEvent, OTX_TIME_INFINITY);
    //DemoLogf("Event signalled\n");
    do {
      result = OtxDrvGetEventData(g_hEventQueue, &eventData); DemoCheckRc(g_hEventQueue, result);

      if (OTX_RESULT_CODE(result) == OTX_S_OK) {
        // We have a driver event
        // Printf the event data.
        // In a real application branch based on the data
        DemoPrintEvent(&eventData);

      }
    } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  }

  return 0;


}



/*******************************************************************************
  @func
    Prints a Driver Event. If HDLC Data, reads the data.
*******************************************************************************/
void DemoPrintEvent(
  OtxEventDataS *pEventData
)
{
  OTX_RESULT otxRc;
  char msgStr[256];

  DemoPrintTimeStamp(pEventData->m_nTimeStamp);

  switch (pEventData->m_eDeviceType) {
    case OTX_DEVICE_LI_T1E1:
    // Event from Li0 or Li1
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
      if (pEventData->m_nParam) {
        DemoLogf("E1/T1[%d] Event: %s (%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam);
      } else {
        DemoLogf("E1/T1[%d] Event: %s\n", pEventData->m_nSourceId, msgStr);
      }
    } else {
      DemoLogf("E1/T1[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_DEVICE_DSP:
    // Non-logical device related event from the DSPs
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    if (OTX_SUCCESS(otxRc)) {
       DemoLogf("DSP[%d] Event: %s (m_nParam=0x%X, m_nRequestId=0x%X)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam, pEventData->m_nRequestId);

    //	DemoLogf("DSP[%d] Event: %s (m_nParam=%d, m_nRequestId=%d)\n", pEventData->m_nSourceId, msgStr, pEventData->m_nParam, pEventData->m_nRequestId);
    } else {
      DemoLogf("DSP[%d] Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
    }
    break;

  case OTX_LDEVICE_HDLC_SENDER:
    otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 256, &msgStr[0]);
    DemoLogf("HdlcSender[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
    break;

  default:
    // Unhandled event
    DemoLogf("Event from an unknown object 0x%X\n", pEventData->m_hDevice);
    break;
  }


}



/*******************************************************************************

  @func

  Transmits MSU frames

  @comm

*******************************************************************************/
void *DemoMsuTransmitFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 j;
  OtxHdlcSS7FisuS msu;
  OtxHdlcSS7FisuS *pFisu;
  OTX_UINT8 *sqHdlcFrame;
  OTX_INT32 cbHdlcFrame;
  OTX_UINT32 nDev;

  nDev = *((OTX_INT32 *) lpParameter);

  g_sqDspHdlcSender[nDev].nByteCount = 0;
  g_sqDspHdlcSender[nDev].nBusyCount = 0;
  g_sqDspHdlcSender[nDev].bActive = OTX_TRUE;
  
  sqHdlcFrame = (OTX_UINT8 *) malloc(300);
  if (!sqHdlcFrame) {
    DemoLogf("DemoMsuTransmitFunction[%d] Out of memory\n", nDev);
    goto TXDONE;
  }

  srand(nDev);

  // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures
  pFisu = &g_sqDspHdlcSender[nDev].tFisu;

  DemoLogf("DemoMsuTransmitFunction[%d] Sending %d MSUs from (Dsp#%d, Li#%d, TS%d)\n", nDev, g_sqDspHdlcSender[nDev].coMsusToSend, g_sqDspHdlcSender[nDev].nDsp, g_sqDspHdlcSender[nDev].nLi, g_sqDspHdlcSender[nDev].nTs);

  g_sqDspHdlcSender[nDev].startTime = clock();

  for (i = 0; i < g_sqDspHdlcSender[nDev].coMsusToSend; i++) {

    if (!g_sqDspHdlcSender[nDev].bActive) goto TXDONE; // Thread aborted

    g_counter++;
    cbHdlcFrame = (rand() % DEMO_MAX_FRAME_LEN) + 6;

    pFisu->fsn++;
    
    memcpy(&msu, pFisu, OTX_HDLC_SS7_CB_FISU);  // Build the first 3 bytes of the MSU
    msu.li = cbHdlcFrame - OTX_HDLC_SS7_CB_FISU;  // Set the Length Indicator
    memcpy(sqHdlcFrame, &msu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
    for (j = OTX_HDLC_SS7_CB_FISU; j < cbHdlcFrame; j++) {     // Add the payload the HDLC frame buffer
      sqHdlcFrame[j] = (OTX_UINT8) rand();
    }
    
    // Send an MSU Frame
    result = OTX_E_BUSY;
    while (result != OTX_S_OK) {
      result = OtxHdlcSS7SendData(g_sqDspHdlcSender[nDev].hHandle, sqHdlcFrame, cbHdlcFrame, pFisu, DEMO_TASK_REF, OTX_TIME_INFINITY); 
      if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
        // Driver IO Request buffer is full. We need to retransmit this frame.
        g_sqDspHdlcSender[nDev].nBusyCount++;
        OsaThreadSleep(50);
      } else {
        DemoCheckRc(g_sqDspHdlcSender[nDev].hHandle, result);
      }
    }

    g_sqDspHdlcSender[nDev].nByteCount += cbHdlcFrame;
    g_sqDspHdlcSender[nDev].coMsusSent++;
    
    // Logging
    if (g_sqDspHdlcSender[nDev].bVerbose) {
      OTX_INT32 k;

      DemoLogf("DemoMsuTransmitFunction[%d] Sent #%d (len=%d) ", nDev, g_sqDspHdlcSender[nDev].coMsusSent, cbHdlcFrame);
      for (k = 0; k < cbHdlcFrame; k++) {
        DemoLogf("%.2X ", sqHdlcFrame[k]);
      }
      DemoLogf("\n");
    }

    OsaThreadSleep((g_sqDspHdlcSender[nDev].nMinSleepTimeBetweenHdlcFrames + cbHdlcFrame) / g_sqDspHdlcSender[nDev].nLoadFactor);

    // Check for events from this device
    {
      OtxEventDataS eventData;
      OtxEventDataS *pEventData;
      OTX_RESULT otxRc;
      char msgStr[128];
      
      do {
        result = OtxDrvGetEventData(g_sqDspHdlcSender[nDev].hEventQueue, &eventData); DemoCheckRc(g_sqDspHdlcSender[nDev].hEventQueue, result);
        
        if (OTX_RESULT_CODE(result) == OTX_S_OK) {
          // We have a driver event
          pEventData = &eventData;
          
          switch (pEventData->m_eDeviceType) {
            
          case OTX_LDEVICE_HDLC_SENDER:
            otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 128, &msgStr[0]);
            DemoLogf("DemoMsuTransmitFunction[%d] Event: %s \n", pEventData->m_nSourceId, msgStr);
            break;
            
          default:
            // Unhandled event
            DemoLogf("DemoMsuTransmitFunction[%d]: Event from an unknown object 0x%X\n", nDev, pEventData->m_hDevice);
            break;
          }
        }
        
      } while (OTX_RESULT_CODE(result) == OTX_S_OK);
    }
  }

TXDONE:

  g_sqDspHdlcSender[nDev].bActive = OTX_FALSE;

  if (sqHdlcFrame) {
    free(sqHdlcFrame);
  }

  g_sqDspHdlcSender[nDev].stopTime = clock();
  
  DemoLogf("DemoMsuTransmitFunction[%d] Thread Exiting (BusyCount=%d, time=%2.1f seconds).\n", nDev, g_sqDspHdlcSender[nDev].nBusyCount, (double)(g_sqDspHdlcSender[nDev].stopTime - g_sqDspHdlcSender[nDev].startTime) / CLOCKS_PER_SEC);
  
  return 0;
}



/*******************************************************************************

  @func

  Receives MSU frames

  @comm

*******************************************************************************/
void *DemoMsuReceiveFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 j;
  clock_t startTime = 0, stopTime;
  OTX_UINT32 nDev;
  OTX_UINT8  *buf;
  OTX_BOOL bErroredFrame = OTX_FALSE;
  
  nDev = *((OTX_INT32 *) lpParameter);
  
  srand(nDev);
  
  g_sqDspHdlcReceiver[nDev].bActive = OTX_TRUE;
  
  buf = (OTX_UINT8 *) malloc(DEMO_HDLC_RX_BUFFER_SIZE);
  if (!buf) {
    DemoLogf("DemoMsuReceiveFunction[%d] Out of memory\n", nDev);
    goto RXDONE;
  }
  
  DemoLogf("DemoMsuReceiveFunction[%d] Starting HDLC Receiver (Dsp#%d, Li#%d, TS%d)\n", nDev, g_sqDspHdlcReceiver[nDev].nDsp, g_sqDspHdlcReceiver[nDev].nLi, g_sqDspHdlcReceiver[nDev].nTs);

  // Discard any events from event queue (if any)
  j = 0;
  do {
    OtxEventDataS eventData;
    result = OtxDrvGetEventData(g_sqDspHdlcReceiver[nDev].hEventQueue, &eventData); DemoCheckRc(g_sqDspHdlcReceiver[nDev].hEventQueue, result);
    if (OTX_RESULT_CODE(result) == OTX_S_OK) j++;
  } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  DemoLogf("DemoMsuReceiveFunction[%d] Discarded %d events from the event queue\n", nDev, j);

  // Discard any data from the Raw Data Receiver
  j = 0;
  do {
    OTX_UINT32 nFrameLen;
    OTX_UINT32 nFrameSeq;
    OtxHdlcFrameEndCodeE nFrameEndCode;
    OTX_DATETIME nTimeStamp;
              
    result = OtxHdlcReadData(g_sqDspHdlcReceiver[nDev].hHandle, buf, DEMO_HDLC_RX_BUFFER_SIZE, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
    if (OTX_RESULT_CODE(result) == OTX_S_OK) j++;
  } while (OTX_RESULT_CODE(result) == OTX_S_OK);
  if (OTX_RESULT_CODE(result) != OTX_S_FALSE) {
    DemoLogf("DemoMsuReceiveFunction[%d] Discarded %d events from the event queue\n", j);
    DemoLogf("RawDataReceiver[%d] Could not flush the HDLC fifo\n", nDev);
    DemoCheckRc(g_sqDspHdlcReceiver[nDev].hHandle, result);
  }
  DemoLogf("DemoMsuReceiveFunction[%d] Discarded %d HDLC frames from the device\n", nDev, j);
  
  startTime = clock();
  while (!bErroredFrame && g_sqDspHdlcReceiver[nDev].coMsusReceived < g_sqDspHdlcReceiver[nDev].coMsusToReceive) {
    
    OtxDrvWaitForSingleEvent(g_sqDspHdlcReceiver[nDev].hNotificationEvent, OTX_TIME_INFINITY);

    //DemoLogf("DemoMsuReceiveFunction[%d] Event signalled\n", nDev);
    
    // Check for events from this device
    {
      OtxEventDataS eventData;
      OtxEventDataS *pEventData;
      OTX_RESULT otxRc;
      char msgStr[128];
      
      do {
        result = OtxDrvGetEventData(g_sqDspHdlcReceiver[nDev].hEventQueue, &eventData); DemoCheckRc(g_sqDspHdlcReceiver[nDev].hEventQueue, result);
        
        if (OTX_RESULT_CODE(result) == OTX_S_OK) {
          // We have a driver event
          pEventData = &eventData;
          
          switch (pEventData->m_eDeviceType) {
            
          case OTX_LDEVICE_HDLC_RECEIVER:
            // Event from HLDC Receiver
            if (pEventData->m_nCode == OTX_HDLC_RECEIVER_EC_COMPLETE_FRAME) {
              OTX_UINT32 nFrameLen;
              OTX_UINT32 nFrameSeq;
              OtxHdlcFrameEndCodeE nFrameEndCode;
              OTX_DATETIME nTimeStamp;
              OtxHdlcSS7FisuS *pFisu;

              assert(nDev == pEventData->m_nSourceId);
              assert(g_sqDspHdlcReceiver[nDev].hHandle == pEventData->m_hDevice);

              otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, DEMO_HDLC_RX_BUFFER_SIZE, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
              while (OTX_RESULT_CODE(otxRc) == OTX_S_OK) {
                
                if (nFrameEndCode == 0) {
                  
                  pFisu = (OtxHdlcSS7FisuS *)&buf[0];
                  
                  if (((OTX_UINT8)pFisu->bsn != g_sqDspHdlcReceiver[nDev].nPrevBsn) && ((OTX_UINT8)pFisu->bsn != ((g_sqDspHdlcReceiver[nDev].nPrevBsn+1) % 128))) {
                    bErroredFrame = OTX_TRUE;
                    DemoLogf("DemoMsuReceiveFunction[%d]: Bsn Error\n", nDev);
                    OtxDrvDisable(pEventData->m_hDevice, OTX_TASK_SYNC, 5000L);
                  }
                  
                  if (nFrameLen == 3) {
                    // FISU
                    // No further check
                    DemoLogf("DemoMsuReceiveFunction[%d]: FISU Received (BSN=%d)\n", nDev, pFisu->bsn);


                  } else {
                    // MSU
                    OTX_UINT32 nExpectedFrameLen = (rand() % DEMO_MAX_FRAME_LEN) + 6;

                    // Verbose logging
                    if (g_sqDspHdlcReceiver[nDev].bVerbose) {
                      OTX_UINT32 k;
                      
                      DemoLogf("DemoMsuReceiveFunction[%d]: Rcvd #%d (len=%d) ", nDev, g_sqDspHdlcReceiver[nDev].coMsusReceived+1, nFrameLen);
                      
                      for (k = 0; k < nFrameLen; k++) {
                        DemoLogf("%.2X ", buf[k]);
                      }
                      DemoLogf("\n");
                    }


                    if (nFrameLen != nExpectedFrameLen) {
                      // Data length check
                      bErroredFrame = OTX_TRUE;
                      DemoLogf("DemoMsuReceiveFunction[%d]: Length Error (%d). Expected %d. Got %d.\n", nDev, nExpectedFrameLen, nFrameLen);
                      OtxDrvDisable(pEventData->m_hDevice, OTX_TASK_SYNC, 5000L);
                      goto RXDONE;
                    } else {
                      // Payload check
                      OTX_UINT8 nExpectedByte;
                      for (i = 3; i < (OTX_INT32)nFrameLen && i < DEMO_HDLC_RX_BUFFER_SIZE; i++) {
                        nExpectedByte = (OTX_UINT8) rand();
                        if (buf[i] != nExpectedByte) {
                          bErroredFrame = OTX_TRUE;
                          DemoLogf("DemoMsuReceiveFunction[%d]: Payload Error. Expected [%d:%.2X]. Got [%d:%.2X].\n", nDev, i, buf[i], i, nExpectedByte);
                          OtxDrvDisable(pEventData->m_hDevice, OTX_TASK_SYNC, 5000L);
                          goto RXDONE;
                        }
                      }
                    }
                    
                    g_sqDspHdlcReceiver[nDev].coMsusReceived++;
                    
                  }
                  
                  g_sqDspHdlcReceiver[nDev].nPrevBsn = pFisu->bsn;
                  
                }
                
                // Read the next frame, if any
                otxRc = OtxHdlcReadData(pEventData->m_hDevice, buf, DEMO_HDLC_RX_BUFFER_SIZE, &nFrameLen, &nFrameSeq, &nFrameEndCode, &nTimeStamp);
              }
              if (OTX_RESULT_CODE(otxRc) != OTX_S_FALSE) {
                DemoLogf("DemoMsuReceiveFunction[%d]: OtxHdlcReadData() was unsuccessful.\n", pEventData->m_nSourceId);
                OtxDrvDisable(pEventData->m_hDevice, OTX_TASK_SYNC, 5000L);
                goto RXDONE;
              }
            } else {
              // Unexpected event from this device
              otxRc = OtxDrvEventCode2String(pEventData->m_hDevice, pEventData->m_nCode, 128, &msgStr[0]);
              if (OTX_SUCCESS(otxRc)) {
                DemoLogf("DemoMsuReceiveFunction[%d]:  Event: %s \n", pEventData->m_nSourceId, msgStr);
              } else {
                DemoLogf("DemoMsuReceiveFunction[%d]:  Invalid Event: %ld \n", pEventData->m_nSourceId, pEventData->m_nCode);
              }
              OtxDrvDisable(pEventData->m_hDevice, OTX_TASK_SYNC, 5000L);
              goto RXDONE;
            }
            break;
            
          default:
            // Unhandled event
            DemoLogf("DemoMsuReceiveFunction[%d]: Event from an unknown object 0x%X\n", nDev, pEventData->m_hDevice);
            OtxDrvDisable(pEventData->m_hDevice, OTX_TASK_SYNC, 5000L);
            goto RXDONE;
            break;
          }
        }
        
        
      } while (OTX_RESULT_CODE(result) == OTX_S_OK);
    }
    
  }
  
RXDONE:
  
  g_sqDspHdlcReceiver[nDev].bActive = OTX_FALSE;
  
  if (buf) {
    free(buf);
  }
  
  stopTime = clock();
  
  DemoLogf("DemoMsuReceiveFunction[%d] Thread Exiting (time=%2.1f seconds).\n", nDev, (double)(stopTime - startTime) / CLOCKS_PER_SEC);
  
  return 0;
}



/*******************************************************************************

  @func

  Transmits one or more MSU frames

  @comm

*******************************************************************************/
/*
void *DemoSs7SimulationFunction(
  void *lpParameter   // thread data
)
{
  OTX_RESULT result;
  OTX_INT32 i;
  OTX_INT32 busyCount = 0;
  clock_t startTime, stopTime;
  OtxHdlcSS7FisuS msu;
  OtxHdlcSS7FisuS fisu;
  OTX_UINT8 sqHdlcFrame[200];
  OTX_INT32 cbHdlcFrame;
  OTX_UINT32 nDev;
  OTX_INT32 coHdlcSenderFrames = DEMO_SS7_SIM_FRAMES;
  OTX_INT32 coMsusSent = 0;
  OTX_INT32 coFisusSent = 0;

  nDev = *((OTX_INT32 *) lpParameter);

  fisu.bib = 1;
  fisu.fib = 1;
  fisu.fsn = 0x7F;
  fisu.bsn = 0x7F;
  fisu.li = OTX_HDLC_SS7_CB_FISU;
  fisu.reserved = 0;
  
  g_bSs7SimulationActive[nDev] = OTX_TRUE;

  memcpy(sqHdlcFrame, &fisu, OTX_HDLC_SS7_CB_FISU);  
  cbHdlcFrame = OTX_HDLC_SS7_CB_FISU;
  
  DemoLogf("Sending a mix of MSU and FISU frames from HDLC Sender #%d  (Dsp#%d, Li#%d)\n", nDev, g_sqDspHdlcSender[nDev].nDsp, g_sqDspHdlcSender[nDev].nLi);
  
  // Send MSU Frame(s)
  startTime = clock();
  while (coMsusSent < coHdlcSenderFrames) {

    result = OtxHdlcSS7SendData(g_sqDspHdlcSender[nDev].hHandle, sqHdlcFrame, cbHdlcFrame, &fisu, DEMO_TASK_REF, OTX_TIME_INFINITY); 
    if (OTX_RESULT_CODE(result) == OTX_E_BUSY) {
      // Driver IO Request buffer is full. We need to retransmit this frame.
      busyCount++;
      OsaThreadSleep(1);
    } else {
      DemoCheckRc(g_sqDspHdlcSender[nDev].hHandle, result);

      // Step BSN for every new frame package (MSU+FISU or FISU+FISU) to send
      fisu.bsn++;

      if (coFisusSent == 0) { 
        OsaThreadSleep(15);
      } else {
        OsaThreadSleep(rand() % 5);
      }

      if (cbHdlcFrame == 3) {
        coFisusSent++;
      } else {
        DemoLogf("%d: Sent MSU# %d\n", nDev, coMsusSent);
        coMsusSent++;
      }

      if ((rand() % 100) < 50) {
        // Prepare an MSU
        cbHdlcFrame = ((coMsusSent%10)*10) + 10;
        memcpy(&msu, &fisu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
        msu.li = cbHdlcFrame - OTX_HDLC_SS7_CB_FISU;  // Set the Length Indicator
        memcpy(sqHdlcFrame, &msu, OTX_HDLC_SS7_CB_FISU); // Put the first 3 bytes of the MSU in the HDLC frame buffer
        for (i = OTX_HDLC_SS7_CB_FISU; i < cbHdlcFrame; i++) {     // Add the payload the HDLC frame buffer
          sqHdlcFrame[i] = (OTX_UINT8) i;
        }


      } else {
        // Prepare a FISU
        memcpy(sqHdlcFrame, &fisu, OTX_HDLC_SS7_CB_FISU);  
        cbHdlcFrame = OTX_HDLC_SS7_CB_FISU;
      }

    }
  }
  stopTime = clock();
  
  DemoLogf("[%d]Done (BusyCount=%d, time=%2.1f seconds) FISUs=%d MSUs=%d.\n", nDev, busyCount, (double)(stopTime - startTime) / CLOCKS_PER_SEC, coFisusSent, coMsusSent);

  return 0;
}
*/





/*******************************************************************************

  @func

  Parse the program's command line arguments and calls upon
  requested actions.

  There are some optional switches that can be used following
  the driver name on the command line. Switches are are
  separated by spaces, and can appear in arbitrary order.
  Switches consist of a hyphen and a letter:

  @comm

*******************************************************************************/
static void DemoParseCmdLineArgs(int argc, char *argv[], DemoEnvT *envP) {
  short  i;

  // Set reasonable defaults before reading from command line
  envP->pszInputStr       = NULL;
  envP->bExitOnError      = OTX_FALSE;
  envP->logFileName       = DEMO_DEFAULT_LOG_FILE;
  envP->eT1E1Mode         = OTX_T1E1_MODE_E1;
  envP->eClkSource        = OTX_CLOCK_SOURCE_LOCAL_0;
  envP->bSs7FisuFiltering = OTX_TRUE;
  envP->boardNo           = DEMO_DEFAULT_BOARD_NO;
  envP->nBoardType        = OTX_DEVICE_THOR_8_PCI_PLUS;
  envP->coHdlcPerLi       = 1; // for now
  envP->coLi              = 4;
  envP->bUseDspDebugger   = OTX_FALSE;
  envP->bUseSenders       = OTX_TRUE;
  envP->bUseReceivers     = OTX_TRUE;
  envP->coDsp             = 4;
  envP->bHdlcBitInversion = OTX_FALSE;
  envP->b56kbpsChannel    = OTX_FALSE;
  envP->eProgramModule    = OTX_DSP_SPM_OTXHDLCEX;
  envP->eT1LineCode       = OTX_T1_LC_B8ZS;
  envP->eT1FrameFormat    = OTX_T1_FF_ESF;
  envP->eE1LineCode       = OTX_E1_LC_HDB3;
  envP->eE1FrameFormat    = OTX_E1_FF_CRC4_MULTIFRAME;

  for (i = 1; i < argc; i++) {        // Scan command line arguments
    if (*(argv[i]+1) == 'Z') {   // Command input string -- case sensitive
      envP->pszInputStr = argv[i]+2;
    } else {    
      OsaStdStrUpr(argv[i]);
      if (*argv[i] == '-' || *argv[i] == '/') {  // Is argument a legal switch?
        if (*(argv[i]+1) == 'L') {   // Log file
          envP->logFileName = argv[i]+2;
        } else if (*(argv[i]+1) == 'Q') {   // Exit on error
          envP->bExitOnError = OTX_TRUE;
        } else if (*(argv[i]+1) == 'S') {   // Senders only
          envP->bUseReceivers = OTX_FALSE;
        } else if (*(argv[i]+1) == 'R') {   // Receivers only
          envP->bUseSenders = OTX_FALSE;
        } else if (*(argv[i]+1) == 'B') {   // Board number
          envP->boardNo = strtoul(argv[i]+2, NULL, 0);
        } else if (*(argv[i]+1) == 'D') {   // DSP Debugger
          envP->bUseDspDebugger = OTX_TRUE;
        } else if (*(argv[i]+1) == 'I') {   // HDLC Bit Inversion
          envP->bHdlcBitInversion = OTX_TRUE;
        } else if (*(argv[i]+1) == '5') {   // Use 56kbps HDLC channel (vs 64kpbs)
          envP->b56kbpsChannel = OTX_TRUE;

        } else if (*(argv[i]+1) == 'K') {   // Clock Source
          if (*(argv[i]+2) == 'I') {   // Internal
            envP->eClkSource = OTX_CLOCK_SOURCE_INTERNAL;
          } else if (*(argv[i]+2) == '0') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_0;
          } else if (*(argv[i]+2) == '1') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_1;
          } else if (*(argv[i]+2) == '2') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_2;
          } else if (*(argv[i]+2) == '3') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_3;
          } else if (*(argv[i]+2) == '4') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_4;
          } else if (*(argv[i]+2) == '5') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_5;
          } else if (*(argv[i]+2) == '6') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_6;
          } else if (*(argv[i]+2) == '7') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_7;
          } else if (*(argv[i]+2) == '8') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_8;
          } else if (*(argv[i]+2) == '9') {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_9;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '0') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_10;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '1') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_11;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '2') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_12;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '3') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_13;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '4') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_14;
          } else if ( (*(argv[i]+2) == '1') && (*(argv[i]+3) == '5') ) {   
            envP->eClkSource = OTX_CLOCK_SOURCE_LOCAL_15;
          } else {
            DemoLogf("Error: Invalid Clock Source command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }

/*
        } else if (*(argv[i]+1) == 'M') {   // Li Mode
          if (*(argv[i]+2) == 'E') {   // E1 
            envP->eT1E1Mode    = OTX_T1E1_MODE_E1;
          } else if (*(argv[i]+2) == 'T') {   // T1
            envP->eT1E1Mode    = OTX_T1E1_MODE_T1;
          } else {
            DemoLogf("Error: Invalid Li Mode (E1/T1) command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
					  exit(1);
          }
*/
        } else if (*(argv[i]+1) == 'P') {   // Dsp Program module
          if (*(argv[i]+2) == 'H') {   // OtxHdlc.out
            envP->eProgramModule = OTX_DSP_SPM_OTXHDLC;
          } else if (*(argv[i]+2) == 'E') {   // OtxHdlcEx.out
            envP->eProgramModule = OTX_DSP_SPM_OTXHDLCEX;
          } else {
            DemoLogf("Error: Invalid DSP Program Module command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }

        } else if (*(argv[i]+1) == 'T') {   // Board Type
          if ( *(argv[i]+2) == 'P') {
            envP->nBoardType  = OTX_DEVICE_THOR_8_PCI_PLUS;
            envP->coLi = 4;
            envP->coDsp = 4;
            envP->coHdlcPerLi = 32; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->nBoardType  = OTX_DEVICE_GIMLE_16_PCI_PLUS;
            envP->coLi = 4;
            envP->coDsp = 4;
            envP->coHdlcPerLi = 32; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            exit(1);
          }
        
        } else if (*(argv[i]+1) == 'F') {   // T1 Frame format
          if ( *(argv[i]+2) == 'E') {
            envP->eT1FrameFormat = OTX_T1_FF_ESF; // ESF
          } else if ( *(argv[i]+2) == '4') {
            envP->eT1FrameFormat = OTX_T1_FF_F4; 
          } else if ( *(argv[i]+2) == 'S') {
            envP->eT1FrameFormat = OTX_T1_FF_F12; // SF
          } else if ( *(argv[i]+2) == '9') {
            envP->eT1FrameFormat = OTX_T1_FF_F72; // SLC96
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'J') {   // T1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eT1LineCode = OTX_T1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'B') {
            envP->eT1LineCode = OTX_T1_LC_B8ZS; // B8ZS
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'G') {   // E1 Frame format
          if ( *(argv[i]+2) == 'D') {
            envP->eE1FrameFormat = OTX_E1_FF_DOUBLE_FRAME; 
          } else if ( *(argv[i]+2) == 'C') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME; 
          } else if ( *(argv[i]+2) == 'G') {
            envP->eE1FrameFormat = OTX_E1_FF_CRC4_MULTIFRAME_G706; 
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else if (*(argv[i]+1) == 'U') {   // E1 Line Code
          if ( *(argv[i]+2) == 'A') {
            envP->eE1LineCode = OTX_E1_LC_AMI; // AMI
          } else if ( *(argv[i]+2) == 'H') {
            envP->eE1LineCode = OTX_E1_LC_HDB3; // HDB3
          } else {
            DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
            DemoHelpPanel();
            OsaConsoleExit();
            exit(1);
          }
          
        } else {
          DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
          DemoHelpPanel();
          OsaConsoleExit();
				  exit(1);
        }
      } else {
        DemoLogf("Error: Unknown command line switch %s\n", argv[i]);
        DemoHelpPanel();
        OsaConsoleExit();
			  exit(1);
      }
    }
  }

}


/*******************************************************************************

  @func

  Displays a help for the command line.

  @comm

*******************************************************************************/
static void DemoHelpPanel()
{
  DemoLogf("\nCommand line switches:\n");
  DemoLogf(" /B<boardNo>   : Board number (Default is %d)\n", DEMO_DEFAULT_BOARD_NO);
  DemoLogf(" /L<fileName>  : Log file name\n");
  DemoLogf(" /T[P|G]       : Board Type used for the demo, where\n");
  DemoLogf("                   P = Use Thor-8-PCI-Plus with Vidar-55x4-ASM Board\n");
  DemoLogf("                   G = Use Gimle-16-PCI-Plus with Vidar-55x4-ASM Board\n");
  //DemoLogf("                   S = Use Sleipnir-1-PCI (DSP version)\n");

  DemoLogf(" /K[I|0]       : Clock Source, where\n");
  DemoLogf("                   I = Internal clock\n");
  DemoLogf(" /ME           : E1 Li mode\n");
  //DemoLogf(" /MT           : T1 Li mode\n");
  DemoLogf(" /D            : DSP Debugger mode. Prompt to start DSP code from Debugger\n");
  DemoLogf(" /I            : Use HDLC Bit Inversion (Default is to not invert the bits)\n");
  DemoLogf(" /5            : Use HDLC 56kbps channel. Only for OTXHDLC.\n");
  DemoLogf(" /S            : Only HDLC Senders (No Receivers).\n");
  DemoLogf(" /R            : Only HDLC Receivers (No Senders). Required for Gimle-16-PCI-Plus\n");
  DemoLogf(" /P[H|E]       : DSP program to load, where\n");
  DemoLogf("                   H = OtxHdlc\n");
  DemoLogf("                   E = OtxHdlcEx (default)\n");
  DemoLogf(" /F[E|S|9]     : T1 Frame Format, where\n");
  DemoLogf("                   E = ESF (default)\n");
  DemoLogf("                   4 = F4\n");
  DemoLogf("                   S = SF [D3/D4] (F12)\n");
  DemoLogf("                   9 = SLC96 (F72)\n");
  DemoLogf(" /J[A|B]       : T1 Line Code, where\n");
  DemoLogf("                   A = AMI\n");
  DemoLogf("                   B = B8ZS (default)\n");
  DemoLogf(" /G[D|C|G]     : E1 Frame Format, where\n");
  DemoLogf("                   D = Doubleframe\n");
  DemoLogf("                   C = CRC4 Multiframe (default)\n");
  DemoLogf("                   G = CRC4 Multiframe (G.706)\n");
  DemoLogf(" /U[A|H]       : E1 Line Code, where\n");
  DemoLogf("                   A = AMI\n");
  DemoLogf("                   H = HDB3 (default)\n");
  DemoLogf(" /Z<str>       : Command string (overrides keyboard input)\n");
  DemoLogf(" /Q            : Exit without prompting on errors\n");
}


/*******************************************************************************

  @func

  Displays a message on the screen and prints it to the logfile.

  @comm

*******************************************************************************/
static void DemoLogf(char *formatStr, ...)
{
  va_list argList;

  OsaMutexLock(&g_demoLogfCritSec);

  va_start(argList, formatStr);

  OsaConsolePrintF(formatStr, argList);
  if (g_pDemoLogFile) vfprintf(g_pDemoLogFile,   formatStr, argList);

  va_end(argList);

  OsaMutexUnlock(&g_demoLogfCritSec);

}


static void DemoCheckRc(IN OTX_HANDLE hDevice, IN OTX_RESULT result)
/*******************************************************************************
  @func
    Check the result from API call. If the result is a failure code then prompt
    the user if he wants to exit.
*******************************************************************************/
{
  if ( OTX_FAILURE(result) ) {

    if (hDevice == OTX_INVALID_HANDLE_VALUE) {
      DemoLogf("ERROR(%#lX) \n", result);
    } else {
      OTX_RESULT otxRc;
      char msgStr[256];
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      OTX_UINT32      eDeviceType = 999;    // Device type for hDevice 
      OTX_UINT32      nDeviceNumber = 999;  // Device number for hDevice
      
      // Get the Device Type for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_TYPE, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_TYPE, &eDeviceType); 
      } 
      
      // Get the Device Number for hHandle
      otxRc = OtxDrvGetAttributeType(hDevice, OTX_ATTR_DEVICE_NO, &eAttribType, &cbAttribValue, &bReadOnly); 
      if (OTX_SUCCESS(otxRc)) {
        assert(eAttribType == OTX_ATTRIB_UINT32);
        assert(cbAttribValue == sizeof(OTX_UINT32));
        OtxDrvGetAttributeValueUint32(hDevice, OTX_ATTR_DEVICE_NO, &nDeviceNumber); 
      }
      
      // Convert the result code to a more descriptive string
      otxRc = OtxDrvResultCode2String(hDevice, result, 256, &msgStr[0]);
      if (OTX_FAILURE(otxRc)) {
        strcpy(msgStr, "Cannot translate this result code.");
      }
      DemoLogf("ERROR(%#lX [%d.%d]) Dev(%X,%d,%d): %s \n", OTX_RESULT_CODE(result), OTX_RESULT_SOURCE_ID(result), OTX_RESULT_LINE_NUM(result), hDevice, (int) eDeviceType, (int)nDeviceNumber, msgStr);
    }

    DemoExitQuestion();
  }
}


/*******************************************************************************
  @func
    Prompts whether to exit the program. Usually called after an error 
    has been reported.

*******************************************************************************/
void DemoExitQuestion()
{
  short c;

  if (g_demoEnv.bExitOnError) {
    c = 'y';
  } else {
    DemoLogf("Do you want to exit (Y/N) ");
    c = OsaConsoleGetCh();
  }
  if (c == 'y' || c == 'Y') {
    // Clean up
    OtxDrvReleaseLib();
    OsaConsoleExit();
    exit(1);
  }
}


/*******************************************************************************

  @func

  Processes User Commands

  @comm

*******************************************************************************/
OTX_INT32 DemoProcessUserCommand(char commandChar)
{
  static OTX_INT32      nCurrentHdlcDev = 1;
  OTX_RESULT result;

  DemoLogf("\nKey %c pressed\n", commandChar);

  switch (commandChar) {

  case '?':
  case 'h':
  case 'H':
    DemoLogf("Cmd  Synopsis\n");
    DemoLogf("===  ========\n");
    DemoLogf(" ?   Display Help text\n");
    DemoLogf(" h   Ditto\n");
    DemoLogf(" q   Quit\n");
    DemoLogf(" l   Query Line Status of the two Lis\n");
    DemoLogf(" s   Display Sender and Receiver thread status\n");
    DemoLogf(" S   Display the SS7 HDLC receive status of the active HDLC Receiver\n");
    DemoLogf(" n   Increment current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
    DemoLogf(" N   Decrement current HDLC Device (currently HdlcDev#%d selected\n", nCurrentHdlcDev);
    DemoLogf(" y   Query HDLC Receiver Fifo diagnostics (for current HDLC device)\n");
    DemoLogf(" v   Toggle verbose logging of currnet HDLC device\n");
    DemoLogf(" k   Get Notification Event Buffer Diagnostics for current DSP\n");
    break;

  case 'v':
    {
          
      g_sqDspHdlcSender[nCurrentHdlcDev].bVerbose = !g_sqDspHdlcSender[nCurrentHdlcDev].bVerbose;
      g_sqDspHdlcReceiver[nCurrentHdlcDev].bVerbose = !g_sqDspHdlcReceiver[nCurrentHdlcDev].bVerbose;
        
      DemoLogf("[%.3d] Sender   Verbose is now %s.\n", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].bVerbose ? "TRUE": "FALSE");
      DemoLogf("[%.3d] Receiver Verbose is now %s.\n", nCurrentHdlcDev, g_sqDspHdlcReceiver[nCurrentHdlcDev].bVerbose ? "TRUE": "FALSE");

    }
    break;


  case 'k':
    {
      OTX_UINT32 nAttribValue;
      OTX_INT32 nDsp = g_sqDspHdlcSender[nCurrentHdlcDev].nDsp;
      DemoLogf("Notification Event Buffer Diagnostics for DSP#%d: (nDev=%d)\n", g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, nCurrentHdlcDev);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_SIZE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Size (in words)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_CURRENT_USAGE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Current usage (in words): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_MAX_USAGE, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Max usage (in words)    : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_hPhysDevs.hDsp[nDsp], OTX_ATTR_DSP_NOTIFICATION_EVENT_BUF_LOST_EVENTS, &nAttribValue); DemoCheckRc(g_hPhysDevs.hDsp[nDsp], result);
      DemoLogf("  Number of lost events   : %d\n", nAttribValue);
    }
    break;

  case 'y':
    {
      OTX_UINT32 nAttribValue;
      DemoLogf("Fifo Diagnostics for HDLC Receiver device#%d:\n", nCurrentHdlcDev);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_SIZE_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Size (in frames)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_SIZE_OCTETS, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Size (in octets)         : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_PEAK_USAGE_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Peak usage (in frames)   : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_PEAK_USAGE_OCTETS, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Peak usage (in octets)   : %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_CURRENT_USAGE_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Current usage (in frames): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_CURRENT_USAGE_OCTETS, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Current usage (in octets): %d\n", nAttribValue);

      result = OtxDrvGetAttributeValueUint32(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, OTX_ATTR_HDLC_RECEIVER_FIFO_LOST_FRAMES, &nAttribValue); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);
      DemoLogf("  Number of lost frames    : %d\n", nAttribValue);
    }
    break;


  case 'S':
    {
      OTX_UINT32  nFisuCount;   // Number of FISUs received
      OTX_UINT32  nLssuCount;   // Number of LSSUs received
      OTX_UINT32  nMsuCount;    // Number of MSUs received
      OtxHdlcSS7FisuS fisu;     // The last received FISU on the pipe.
      DemoLogf("SS7 Status for HdlcReceiver #%d \n", nCurrentHdlcDev);

      result = OtxHdlcSS7GetReceiveStatus(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, &nFisuCount, &nLssuCount, &nMsuCount, &fisu); DemoCheckRc(g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle, result);

      if (OTX_SUCCESS(result)) {
        DemoLogf("FISUs received: %lu\n", nFisuCount);
        DemoLogf("LSSUs received: %lu\n", nLssuCount);
        DemoLogf("MSUs  received: %lu\n", nMsuCount);
        DemoLogf("Last FISU received: bsn = %d, bib = %d, fsn = %d, fib = %d, li = %d\n",
               fisu.bsn, fisu.bib, fisu.fsn, fisu.fib, fisu.li);
      }

    }
    break;

  case 'l':
    {
      OtxT1E1Layer1StatusE eLayer1Status;
      OTX_INT32 i;

      for (i = 0; i < g_demoEnv.coLi; i++) {
        DemoLogf("Line status of Li #%d is ", i);
        result = OtxT1E1GetLayer1Status(g_hPhysDevs.hLi[i], &eLayer1Status); DemoCheckRc(g_hPhysDevs.hLi[i], result);
        DemoLogf("%s\n", eLayer1Status == OTX_T1E1_LAYER1_OK ? "OK" : "DOWN");
      }

    }
    break;

  case 'F':
    {
      OTX_HANDLE hHandle;
      OTX_UINT32 filterMask;

      hHandle = g_sqDspHdlcReceiver[nCurrentHdlcDev].hHandle;

      result = OtxHdlcSS7GetFilter(hHandle, &filterMask); DemoCheckRc(hHandle, result);
      if (filterMask & OTX_HDLC_SS7_FILTER_FISU_ALL) {
        filterMask &= ~OTX_HDLC_SS7_FILTER_FISU_ALL;
      } else {
        filterMask |= OTX_HDLC_SS7_FILTER_FISU_ALL;
      }

      DemoLogf("Setting SS7 filter mask to %.4X for HDLC Receiver #%d (Dsp#%d, Li#%d)...", filterMask, nCurrentHdlcDev, g_sqDspHdlcReceiver[nCurrentHdlcDev].nDsp, g_sqDspHdlcReceiver[nCurrentHdlcDev].nLi);

      result = OtxHdlcSS7SetFilter(hHandle, filterMask); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;
/*
  case 'f':
    {
      OTX_HANDLE hHandle;

      hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;
      DemoLogf("Start Sending FISUs (fsn=%d) continously from HDLC Sender #%d (Dsp#%d, Li#%d)...", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);

      g_sqDspHdlcSender[nCurrentHdlcDev].tFisu.fsn++;
      result = OtxHdlcStartRepeatData(hHandle, (OTX_UINT8 *) &g_sqDspHdlcSender[nCurrentHdlcDev].tFisu, OTX_HDLC_SS7_CB_FISU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;
*/
/*
  case 'i':
    {
      OTX_HANDLE hHandle;

      hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;

      // Note: The following action merely demonstrates the use of the OTX SS7 API functions. It does not necessarily demonstrate proper SS7 protocol procedures
      g_tLssu.fsn = g_tFisu0.fsn;
      g_tLssu.sf = g_tLssu.sf + 1; 
      g_tLssu.li = 1; // Short LSSU

      DemoLogf("Start Sending LSSUs (fsn=%d, sf=%d) continously from HDLC Sender #%d (Dsp#%d, Li#%d)...", g_tLssu.fsn, g_tLssu.sf, nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);

      result = OtxHdlcStartRepeatData(hHandle, (OTX_UINT8 *) &g_tLssu, OTX_HDLC_SS7_CB_SHORT_LSSU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;
*/
/*
  case 't':
    {
  	  OsaThreadStartEx(DemoMsuReceiveFunction,  (void *) &g_sqDspHdlcReceiver[nCurrentHdlcDev].nDev);
  	  OsaThreadStartEx(DemoMsuTransmitFunction, (void *) &g_sqDspHdlcSender[nCurrentHdlcDev].nDev);
    }
    break;

  case 'T':
    {
      OTX_INT32  i;
      
      for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
        if (g_sqDspHdlcSender[i].nTs == 0) continue; // Skip TS0
        OsaThreadStartEx(DemoMsuReceiveFunction,  (void *) &g_sqDspHdlcReceiver[i].nDev);
        OsaThreadStartEx(DemoMsuTransmitFunction, (void *) &g_sqDspHdlcSender[i].nDev);
      }
    }
    break;
*/
/*
  case 'r':
    {
      OTX_UINT8 *pRepeatFrame = (OTX_UINT8 *)"\x10\x11\x12\x13\x14";
      OTX_INT32 nMsgLen = 5;
      OTX_HANDLE hHandle;

      hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;

      DemoLogf("Start Repeat Data Frames from HDLC Sender #%d (Dsp#%d, Li#%d)...", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);

      result = OtxHdlcStartRepeatData(hHandle, pRepeatFrame, nMsgLen, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;

  case 'R':
    {
      OTX_HANDLE hHandle;

      hHandle = g_sqDspHdlcSender[nCurrentHdlcDev].hHandle;

      DemoLogf("Stop Repeat Data Frames from HDLC Sender #%d (Dsp#%d, Li#%d)...", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);

      result = OtxHdlcStopRepeatData(hHandle); DemoCheckRc(hHandle, result);

      DemoLogf("Done.\n");
    }
    break;
*/


  case '>':
    {
      // Increase load factor for all Sender devices
      OTX_INT32 nLoadFactor;
      OTX_INT32 i;

      nLoadFactor = g_sqDspHdlcSender[nCurrentHdlcDev].nLoadFactor;
      nLoadFactor++;

      for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
        if (g_sqDspHdlcSender[i].nTs == 0) continue;
        g_sqDspHdlcSender[i].nLoadFactor = nLoadFactor;
      }

      DemoLogf("Load factor increased to %d\n", nLoadFactor);
      
    }
    break;
    
  case '<':
    {
      // Decrease load factor for all Sender devices
      OTX_INT32 nLoadFactor;
      OTX_INT32 i;

      nLoadFactor = g_sqDspHdlcSender[nCurrentHdlcDev].nLoadFactor;
      nLoadFactor--;
      if (nLoadFactor < 1) {
        nLoadFactor = 1;
      }

      for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
        if (g_sqDspHdlcSender[i].nTs == 0) continue;
        g_sqDspHdlcSender[i].nLoadFactor = nLoadFactor;
      }

      DemoLogf("Load factor decreased to %d\n", nLoadFactor);
      
    }
    break;
    
    

  case 'n':
    {
      // Select current nCurrentHdlcDev
      nCurrentHdlcDev++;

      if (g_sqDspHdlcSender[nCurrentHdlcDev].nTs == 0) {
        nCurrentHdlcDev++; // Skip TS0 (Sync)
      } 

      if (nCurrentHdlcDev > DEMO_NO_OF_DEVS-1) {
        // Wrap 
        nCurrentHdlcDev = 1;
      }
      
      DemoLogf("DSP HDLC Sender#%d (Dsp#%d, Li#%d) Selected \n", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
      
    }
    break;
    
    
  case 'N':
    {
      // Select current nCurrentHdlcDev
      nCurrentHdlcDev--;

      if (g_sqDspHdlcSender[nCurrentHdlcDev].nTs == 0) {
        nCurrentHdlcDev--; // Skip TS0 (Sync)
      } 
      if (nCurrentHdlcDev < 0) {
        // Wrap
        nCurrentHdlcDev = DEMO_NO_OF_DEVS-1;
      }
      
      DemoLogf("DSP HDLC Sender#%d (Dsp#%d, Li#%d) Selected \n", nCurrentHdlcDev, g_sqDspHdlcSender[nCurrentHdlcDev].nDsp, g_sqDspHdlcSender[nCurrentHdlcDev].nLi);
    }
    break;

  case 'I':
    {
      OTX_UINT32 readVal;
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      OTX_INT32 nCurrentLi = 0;

      DemoLogf("Read from an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);

      OsaConsoleGetStr(strBuf); 
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvReadReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, &readVal); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Li%d Address 0x%X = %.2X\n", nCurrentLi, nAddress, (OTX_UINT8)readVal);
    }
    break;

  case 's':
    {
      OTX_INT32  i;
      double load;
      double bytessent;
      double timeinterval;
      static clock_t startTime[DEMO_NO_OF_DEVS];
      static OTX_INT32 nBusyCount[DEMO_NO_OF_DEVS];
      static OTX_INT32 nByteCount[DEMO_NO_OF_DEVS];
      static OTX_INT32 coMsusSent[DEMO_NO_OF_DEVS];
      
      for (i = 0; i < DEMO_NO_OF_DEVS; i++) {

        if (g_sqDspHdlcSender[i].nTs == 0) continue;

        if (g_sqDspHdlcSender[i].bActive) {
          if (startTime[i] == 0) {
            timeinterval = (double)((clock() - g_sqDspHdlcSender[i].startTime)) / CLOCKS_PER_SEC;
          } else {
            timeinterval = (double)((clock() - startTime[i])) / CLOCKS_PER_SEC;
          }
        } else {
          timeinterval = (double)((g_sqDspHdlcSender[i].stopTime - g_sqDspHdlcSender[i].startTime)) / CLOCKS_PER_SEC;
        }

        bytessent = (double)((g_sqDspHdlcSender[i].nByteCount-nByteCount[i]) + (g_sqDspHdlcSender[i].coMsusSent - coMsusSent[i]) * 3 /* CRC + Flag */);

        load = bytessent / (timeinterval * 8000);  /* 8000 bytes per second on a 64kbps pipe */

        DemoLogf("[%.3d] ", i);

        if (g_demoEnv.bUseSenders) {
          DemoLogf(" Tx:%s MSU:%d Busy:%1.2f Byte:%d Load:%2.2f time:%2.2f",
          g_sqDspHdlcSender[i].bActive ? "run" : "end",  
          g_sqDspHdlcSender[i].coMsusSent,  
          (double)(g_sqDspHdlcSender[i].coMsusSent-coMsusSent[i]) / (double)((g_sqDspHdlcSender[i].nBusyCount - nBusyCount[i]) + (g_sqDspHdlcSender[i].coMsusSent - coMsusSent[i])),
          g_sqDspHdlcSender[i].nByteCount,
          load,
          timeinterval
          );
        }

        startTime[i] = clock();
        nBusyCount[i] = g_sqDspHdlcSender[i].nBusyCount;
        nByteCount[i] = g_sqDspHdlcSender[i].nByteCount;
        coMsusSent[i] = g_sqDspHdlcSender[i].coMsusSent;

        if (g_demoEnv.bUseReceivers) {
          DemoLogf(" Rx:%s MSU:%d",
          g_sqDspHdlcReceiver[i].bActive ? "run" : "end",  
          g_sqDspHdlcReceiver[i].coMsusReceived
          );
        }

        DemoLogf("\n");

      }
      DemoLogf("\n");

    }
    break;


  case 'a':
    {
      OTX_INT32  i;
      
      DemoLogf("Aborting all Sender threads...\n");
      for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
        if (g_sqDspHdlcSender[i].nTs == 0) continue;
        g_sqDspHdlcSender[i].bActive = OTX_FALSE;
      }
      DemoLogf("Done\n");
      
    }
    break;


  case 'O':
    {
      char strBuf[12];
      static OTX_UINT32 nAddress = 0;
      static OTX_UINT32 nValue = 0;
      OTX_INT32 nCurrentLi = 0;

      DemoLogf("Write to an 8-bit register in Li%d \n", nCurrentLi);
      DemoLogf("Which Address (0x%X): ", nAddress);
      OsaConsoleGetStr(strBuf);  
      if (*strBuf) {
        nAddress = strtoul(strBuf, NULL, 0);
      }
      DemoLogf("Which Value (0x%X): ", nValue);
      OsaConsoleGetStr(strBuf);   
      if (*strBuf) {
        nValue = strtoul(strBuf, NULL, 0);
      }

      result = OtxDrvWriteReg8(g_hPhysDevs.hLi[nCurrentLi], nAddress, nValue); DemoCheckRc(g_hPhysDevs.hLi[nCurrentLi], result);

      DemoLogf("Wrote %.2X to Li%d Address 0x%X\n", (OTX_UINT8)nValue, nCurrentLi, nAddress);
    }
    break;


    
        
  case '~':
    {
      char strBuf[12];
      OTX_INT32 nSleepTime;

      printf("Enter milliseconds to sleep :");
      OsaConsoleGetStr(strBuf);

      if (strBuf[0]) {
        nSleepTime = strtol(strBuf, NULL, 0);
        printf("Sleeping %d milliseconds\n", nSleepTime);
        OsaThreadSleep(nSleepTime);
      }
    }
    break;

  case 'q':
  case 'Q':
    {
      OTX_INT32  i;
      
      if (g_demoEnv.bUseSenders) {
        DemoLogf("Aborting all Sender threads\n");
        for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
          if (g_sqDspHdlcSender[i].nTs == 0) continue;
          g_sqDspHdlcSender[i].bActive = OTX_FALSE;
        }

        OsaThreadSleep(500);
      }

      DemoLogf("Exiting...\n");
      
      return 1;
    }
    break;

  default:
    DemoLogf("Unknown command. Press '?' for help, 'q' to quit\n");
    break;
  }

  return 0;
}


/*******************************************************************************
*******************************************************************************/
int main(int argc, char *argv[])
{
  OTX_INT32 exitFlag = 0;
  OTX_RESULT result;  // OTX Standard Return Code
  char commandChar;

  DemoInitializeGlobalData();

	OsaConsoleInit();

  OsaMutexInit(&g_demoLogfCritSec);

  DemoLogf("T8pSs7, %s, Odin TeleSystems Inc.\n", __DATE__);
  DemoLogf("HDLC Example program.\n");
  DemoParseCmdLineArgs(argc, argv, &g_demoEnv);

  if (g_demoEnv.pszInputStr) {
    OsaConsoleInput(g_demoEnv.pszInputStr);
  }

  if ((g_pDemoLogFile = fopen(g_demoEnv.logFileName, "wt")) == NULL) {
    printf("Couldn't open logfile %s\n", g_demoEnv.logFileName);
    OsaConsoleExit();
		exit(-1);
  }

  DemoLogf("Log file is %s\n", g_demoEnv.logFileName);
  DemoLogf("Board number %d\n", g_demoEnv.boardNo);

  {
    char szClockSource[64];

    switch (g_demoEnv.eClkSource) {
    case OTX_CLOCK_SOURCE_INTERNAL:
      strcpy(szClockSource, "Internal");
      break;
    case OTX_CLOCK_SOURCE_LOCAL_0:
    case OTX_CLOCK_SOURCE_LOCAL_1:
    case OTX_CLOCK_SOURCE_LOCAL_2:
    case OTX_CLOCK_SOURCE_LOCAL_3:
    case OTX_CLOCK_SOURCE_LOCAL_4:
    case OTX_CLOCK_SOURCE_LOCAL_5:
    case OTX_CLOCK_SOURCE_LOCAL_6:
    case OTX_CLOCK_SOURCE_LOCAL_7:
    case OTX_CLOCK_SOURCE_LOCAL_8:
    case OTX_CLOCK_SOURCE_LOCAL_9:
    case OTX_CLOCK_SOURCE_LOCAL_10:
    case OTX_CLOCK_SOURCE_LOCAL_11:
    case OTX_CLOCK_SOURCE_LOCAL_12:
    case OTX_CLOCK_SOURCE_LOCAL_13:
    case OTX_CLOCK_SOURCE_LOCAL_14:
    case OTX_CLOCK_SOURCE_LOCAL_15:
      sprintf(szClockSource, "Line Interface %d", (int)g_demoEnv.eClkSource - (int)OTX_CLOCK_SOURCE_LOCAL_0);
      break;
    default:
      strcpy(szClockSource, "Unknown");
      break;
    }

    DemoLogf("Clock Source: %s\n", szClockSource);
  }


  // Query Library revision
  {
    char szVersionInfo[255];
    OTX_BOOL bIsStaticLinkedLibrary;
    char szLibFileName[64];
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_REV_STR, szVersionInfo, 255); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueBool(OTX_LIB_HANDLE, OTX_ATTR_LIB_TYPE_STATIC, &bIsStaticLinkedLibrary); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    result = OtxDrvGetAttributeValueSz(OTX_LIB_HANDLE, OTX_ATTR_LIB_FILE_NAME, szLibFileName, 64); DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    
    DemoLogf("Library version %s (filename %s, %s linked)\n", szVersionInfo, szLibFileName, bIsStaticLinkedLibrary ? "static":"dynamic");
  }


  // Connect to the Driver
  DemoLogf("Connecting lib...\n");
  result = OtxDrvConnectLib(); 
  switch (OTX_RESULT_CODE(result)) {
  case OTX_RES_HANDLE:
    DemoLogf("Error: Could not connect to the OTX library.\nPossible cause: OTX Driver not loaded\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  case OTX_RES_NOT_SUPPORTED:
    DemoLogf("Error: Mismatch between OTX Driver revision and OTX Library revision.\nRevisions must match.\n");
    OsaConsoleExit();
    OtxDrvReleaseLib();
    exit(1);
    break;
  default:
    DemoCheckRc(OTX_INVALID_HANDLE_VALUE, result);
    break;
  }

  // Create Event Queues for Driver --> Application Communications
  DemoLogf("Creating event queue...\n");
  DemoCreateEventQueue();
  DemoLogf("Created event queue\n");

  // Connect (get handles) to the physical devices on the Thor-2-PCI board
  DemoLogf("Opening physical devices...\n");
  DemoOpenPhysicalDevices();
  DemoLogf("Opened physical devices\n");

  // Configure E1/T1 devices
  DemoLogf("Configuring E1/T1 devices...\n");
  DemoConfigureE1T1Devices();
  DemoLogf("Configured E1/T1 devices\n");

  // Enable all physical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling physical devices...\n");
  DemoEnablePhysicalDevices();

  // Set the clock source 
  result = OtxBrdSetClocks(g_hPhysDevs.hThorController0, OTX_CLOCK_MODE_LOCAL, g_demoEnv.eClkSource); DemoCheckRc(g_hPhysDevs.hThorController0, result);

  // Create and get handles to the Logical devices needed by this application
  DemoLogf("Creating logical devices...\n");
  DemoCreateLogicalDevices();
  DemoLogf("Created logical devices\n");

  // Create and get handles to the Pipe devices needed by this application
  DemoLogf("Connecting devices...\n");
  DemoConnectDevices();

  // Initialized the Connected Logical Devices
  DemoLogf("Initializing devices...\n");
  DemoInitializeLogicalDevices();

  // Enable all logical devices. Their device state should now be OTX_DEVICE_WORKING
  DemoLogf("Enabling logical devices...\n");
  DemoEnableLogicalDevices();

  // Read Serial Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszSerialNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszSerialNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_SERIAL_NUMBER, pszSerialNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Serial Number: %s\n", pszSerialNumber);
    free(pszSerialNumber);
  }

  // Read Build Number from the board
  {
    OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
    OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
    OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
    char *pszBrdBuildNumber;

    result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    assert(eAttribType == OTX_ATTRIB_SZ);
    assert(bReadOnly == OTX_TRUE);
    pszBrdBuildNumber = (char *) malloc(cbAttribValue * sizeof(char));
    result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_BUILD, pszBrdBuildNumber, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
    DemoLogf("Board Build Number: %s\n", pszBrdBuildNumber);
    free(pszBrdBuildNumber);
  }

  if (g_demoEnv.nBoardType != OTX_DEVICE_THOR_PCMCIA) {
    // Read CPLD Build Numbers from the board
    {
      OtxAttribTypeE  eAttribType;    // Type of the attribute, if attribute exists
      OTX_UINT32      cbAttribValue;  // Size in bytes of attribute value (Get value with <f OtxDrvGetAttributeValue>)
      OTX_BOOL        bReadOnly;      // Is attribute writable or read-only
      char *pszCpldBuildNumbers;
      
      result = OtxDrvGetAttributeType(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, &eAttribType, &cbAttribValue, &bReadOnly); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      assert(eAttribType == OTX_ATTRIB_SZ);
      assert(bReadOnly == OTX_TRUE);
      pszCpldBuildNumbers = (char *) malloc(cbAttribValue * sizeof(char));
      result = OtxDrvGetAttributeValueSz(g_hPhysDevs.hThorController0, OTX_ATTR_BRD_CPLD_BUILD, pszCpldBuildNumbers, cbAttribValue); DemoCheckRc(g_hPhysDevs.hThorController0, result);
      DemoLogf("CPLD Build:\n%s", pszCpldBuildNumbers);
      free(pszCpldBuildNumbers);
    }
  }
  
  // Print events from the driver
	OsaThreadStart(DemoIdleFunction);
  OsaThreadSleep(1000);


  DemoLogf("\nPress any to start HDLC threads...");
  commandChar = OsaConsoleGetCh();

  {
    OTX_INT32  i;

    if (g_demoEnv.bUseSenders) {
      // Start sending FISUs on all devices
      for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
        if (g_sqDspHdlcSender[i].nTs == 0) continue; // Skip TS0
        g_sqDspHdlcSender[i].tFisu.fsn++;
        result = OtxHdlcStartRepeatData(g_sqDspHdlcSender[i].hHandle, (OTX_UINT8 *) &g_sqDspHdlcSender[i].tFisu, OTX_HDLC_SS7_CB_FISU, DEMO_TASK_REF, OTX_TIME_INFINITY); DemoCheckRc(g_sqDspHdlcSender[i].hHandle, result);
      }
      
      OsaThreadSleep(300);
    }

    if (g_demoEnv.bUseReceivers) {
      // Start Receivers
      for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
        if (g_sqDspHdlcReceiver[i].nTs == 0) continue; // Skip TS0
        OsaThreadStartEx(DemoMsuReceiveFunction,  (void *) &g_sqDspHdlcReceiver[i].nDev);
      }
      
      OsaThreadSleep(300);
    }

    if (g_demoEnv.bUseSenders) {
      // Start Senders
      for (i = 0; i < DEMO_NO_OF_DEVS; i++) {
        if (g_sqDspHdlcSender[i].nTs == 0) continue; // Skip TS0
        OsaThreadStartEx(DemoMsuTransmitFunction, (void *) &g_sqDspHdlcSender[i].nDev);
      }
    }

  }

  DemoLogf("\nPress 'h' for help\n");

#if 1
  while (!exitFlag) {
    commandChar = OsaConsoleGetCh();
    exitFlag = DemoProcessUserCommand(commandChar);
  }
#else // Just for testing. Exiting automatically after 300ms
  OsaThreadSleep(300);
#endif

  // Clean up
  OtxDrvReleaseLib();
	OsaConsoleExit();

  return 0;
}

